#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2014-2016 OpenMarket Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
import heapq
import logging
from collections import ChainMap, defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Mapping,
    Optional,
    Sequence,
)

import attr
from immutabledict import immutabledict
from prometheus_client import Counter, Histogram

from synapse.api.constants import EventTypes
from synapse.api.room_versions import KNOWN_ROOM_VERSIONS, StateResolutionVersions
from synapse.events import EventBase
from synapse.events.snapshot import (
    EventContext,
    UnpersistedEventContext,
    UnpersistedEventContextBase,
)
from synapse.logging.context import ContextResourceUsage
from synapse.logging.opentracing import tag_args, trace
from synapse.metrics import SERVER_NAME_LABEL
from synapse.replication.http.state import ReplicationUpdateCurrentStateRestServlet
from synapse.state import v1, v2
from synapse.storage.databases.main.event_federation import StateDifference
from synapse.storage.databases.main.events_worker import EventRedactBehaviour
from synapse.types import StateMap, StrCollection
from synapse.types.state import StateFilter
from synapse.util.async_helpers import Linearizer
from synapse.util.caches.expiringcache import ExpiringCache
from synapse.util.duration import Duration
from synapse.util.metrics import Measure, measure_func
from synapse.util.stringutils import shortstr

if TYPE_CHECKING:
    from synapse.server import HomeServer
    from synapse.storage.controllers import StateStorageController
    from synapse.storage.databases.main import DataStore
    from synapse.storage.databases.state.deletion import StateDeletionDataStore

logger = logging.getLogger(__name__)
metrics_logger = logging.getLogger("synapse.state.metrics")

# Metrics for number of state groups involved in a resolution.
state_groups_histogram = Histogram(
    "synapse_state_number_state_groups_in_resolution",
    "Number of state groups used when performing a state resolution",
    labelnames=[SERVER_NAME_LABEL],
    buckets=(1, 2, 3, 5, 7, 10, 15, 20, 50, 100, 200, 500, "+Inf"),
)


EVICTION_TIMEOUT_SECONDS = 60 * 60


_NEXT_STATE_ID = 1

CREATE_KEY = (EventTypes.Create, "")
POWER_KEY = (EventTypes.PowerLevels, "")


def _gen_state_id() -> str:
    global _NEXT_STATE_ID
    s = "X%d" % (_NEXT_STATE_ID,)
    _NEXT_STATE_ID += 1
    return s


class _StateCacheEntry:
    __slots__ = ["_state", "state_group", "prev_group", "delta_ids"]

    def __init__(
        self,
        state: StateMap[str] | None,
        state_group: int | None,
        prev_group: int | None = None,
        delta_ids: StateMap[str] | None = None,
    ):
        if state is None and state_group is None and prev_group is None:
            raise Exception("One of state, state_group or prev_group must be not None")

        if prev_group is not None and delta_ids is None:
            raise Exception("If prev_group is set so must delta_ids")

        # A map from (type, state_key) to event_id.
        #
        # This can be None if we have a `state_group` (as then we can fetch the
        # state from the DB.)
        self._state: StateMap[str] | None = (
            immutabledict(state) if state is not None else None
        )

        # the ID of a state group if one and only one is involved.
        # otherwise, None otherwise?
        self.state_group = state_group

        self.prev_group = prev_group
        self.delta_ids: StateMap[str] | None = (
            immutabledict(delta_ids) if delta_ids is not None else None
        )

    async def get_state(
        self,
        state_storage: "StateStorageController",
        state_filter: Optional["StateFilter"] = None,
    ) -> StateMap[str]:
        """Get the state map for this entry, either from the in-memory state or
        looking up the state group in the DB.
        """

        if self._state is not None:
            return self._state

        if self.state_group is not None:
            return await state_storage.get_state_ids_for_group(
                self.state_group, state_filter
            )

        assert self.prev_group is not None and self.delta_ids is not None

        prev_state = await state_storage.get_state_ids_for_group(
            self.prev_group, state_filter
        )

        # ChainMap expects MutableMapping, but since we're using it immutably
        # its safe to give it immutable maps.
        return ChainMap(self.delta_ids, prev_state)  # type: ignore[arg-type]

    def set_state_group(self, state_group: int) -> None:
        """Update the state group assigned to this state (e.g. after we've
        persisted it).

        Note: this will cause the cache entry to drop any stored state.
        """

        self.state_group = state_group

        # We clear out the state as we know longer need to explicitly keep it in
        # the `state_cache` (as the store state group cache will do that).
        self._state = None

    def __len__(self) -> int:
        # The len should be used to estimate how large this cache entry is, for
        # cache eviction purposes. This is why it's fine to return 1 if we're
        # not storing any state.

        length = 0

        if self._state:
            length += len(self._state)

        if self.delta_ids:
            length += len(self.delta_ids)

        return length or 1  # Make sure its not 0.


class StateHandler:
    """Fetches bits of state from the stores, and does state resolution
    where necessary
    """

    def __init__(self, hs: "HomeServer"):
        self.server_name = hs.hostname  # nb must be called this for @measure_func
        self.clock = hs.get_clock()  # nb must be called this for @measure_func
        self.store = hs.get_datastores().main
        self._state_storage_controller = hs.get_storage_controllers().state
        self.hs = hs
        self._state_resolution_handler = hs.get_state_resolution_handler()
        self._storage_controllers = hs.get_storage_controllers()
        self._events_shard_config = hs.config.worker.events_shard_config
        self._instance_name = hs.get_instance_name()
        self._state_store = hs.get_datastores().state
        self._state_deletion_store = hs.get_datastores().state_deletion

        self._update_current_state_client = (
            ReplicationUpdateCurrentStateRestServlet.make_client(hs)
        )

    async def compute_state_after_events(
        self,
        room_id: str,
        event_ids: StrCollection,
        state_filter: StateFilter | None = None,
        await_full_state: bool = True,
    ) -> StateMap[str]:
        """Fetch the state after each of the given event IDs. Resolve them and return.

        This is typically used where `event_ids` is a collection of forward extremities
        in a room, intended to become the `prev_events` of a new event E. If so, the
        return value of this function represents the state before E.

        Args:
            room_id: the room_id containing the given events.
            event_ids: the events whose state should be fetched and resolved.
            await_full_state: if `True`, will block if we do not yet have complete state
                at these events and `state_filter` is not satisfied by partial state.
                Defaults to `True`.

        Returns:
            the state dict (a mapping from (event_type, state_key) -> event_id) which
            holds the resolution of the states after the given event IDs.
        """
        logger.debug("calling resolve_state_groups from compute_state_after_events")
        if (
            await_full_state
            and state_filter
            and not state_filter.must_await_full_state(self.hs.is_mine_id)
        ):
            await_full_state = False
        ret = await self.resolve_state_groups_for_events(
            room_id, event_ids, await_full_state
        )
        return await ret.get_state(self._state_storage_controller, state_filter)

    async def get_current_user_ids_in_room(
        self, room_id: str, latest_event_ids: StrCollection
    ) -> set[str]:
        """
        Get the users IDs who are currently in a room.

        Note: This is much slower than using the equivalent method
        `DataStore.get_users_in_room` or `DataStore.get_users_in_room_with_profiles`,
        so this should only be used when wanting the users at a particular point
        in the room.

        Args:
            room_id: The ID of the room.
            latest_event_ids: Precomputed list of latest event IDs. Will be computed if None.
        Returns:
            Set of user IDs in the room.
        """

        assert latest_event_ids is not None

        logger.debug("calling resolve_state_groups from get_current_user_ids_in_room")
        entry = await self.resolve_state_groups_for_events(room_id, latest_event_ids)
        state = await entry.get_state(self._state_storage_controller, StateFilter.all())
        return await self.store.get_joined_user_ids_from_state(room_id, state)

    async def get_hosts_in_room_at_events(
        self, room_id: str, event_ids: StrCollection
    ) -> frozenset[str]:
        """Get the hosts that were in a room at the given event ids

        Args:
            room_id:
            event_ids:

        Returns:
            The hosts in the room at the given events
        """
        entry = await self.resolve_state_groups_for_events(room_id, event_ids)
        return await self._state_storage_controller.get_joined_hosts(room_id, entry)

    @trace
    @tag_args
    async def calculate_context_info(
        self,
        event: EventBase,
        state_ids_before_event: StateMap[str] | None = None,
        partial_state: bool | None = None,
        state_group_before_event: int | None = None,
    ) -> UnpersistedEventContextBase:
        """
        Calulates the contents of an unpersisted event context, other than the current
        state group (which is either provided or calculated when the event context is persisted)

        state_ids_before_event:
            The event ids of the full state before the event if
            it can't be calculated from existing events. This is normally
            only specified when receiving an event from federation where we
            don't have the prev events, e.g. when backfilling or when the event
            is being created for batch persisting.
        partial_state:
            `True` if `state_ids_before_event` is partial and omits non-critical
            membership events.
            `False` if `state_ids_before_event` is the full state.
            `None` when `state_ids_before_event` is not provided. In this case, the
            flag will be calculated based on `event`'s prev events.
        state_group_before_event:
            the current state group at the time of event, if known
        Returns:
            The event context.

        Raises:
            RuntimeError if `state_ids_before_event` is not provided and one or more
                prev events are missing or outliers.
        """
        assert not event.internal_metadata.is_outlier()

        #
        # first of all, figure out the state before the event, unless we
        # already have it.
        #
        if state_ids_before_event:
            # if we're given the state before the event, then we use that
            state_group_before_event_prev_group = None
            deltas_to_state_group_before_event = None

            # the partial_state flag must be provided
            assert partial_state is not None
        else:
            # otherwise, we'll need to resolve the state across the prev_events.

            # partial_state should not be set explicitly in this case:
            # we work it out dynamically
            assert partial_state is None

            # if any of the prev-events have partial state, so do we.
            # (This is slightly racy - the prev-events might get fixed up before we use
            # their states - but I don't think that really matters; it just means we
            # might redundantly recalculate the state for this event later.)
            prev_event_ids = event.prev_event_ids()
            incomplete_prev_events = await self.store.get_partial_state_events(
                prev_event_ids
            )
            partial_state = any(incomplete_prev_events.values())
            if partial_state:
                logger.debug(
                    "New/incoming event %s refers to prev_events %s with partial state",
                    event.event_id,
                    [k for (k, v) in incomplete_prev_events.items() if v],
                )

            logger.debug("calling resolve_state_groups from compute_event_context")
            # we've already taken into account partial state, so no need to wait for
            # complete state here.

            entry = await self.resolve_state_groups_for_events(
                event.room_id,
                event.prev_event_ids(),
                await_full_state=False,
            )

            # Ensure we still have the state groups we're relying on, and bump
            # their usage time to avoid them being deleted from under us.
            if entry.state_group:
                missing_state_group = await self._state_deletion_store.check_state_groups_and_bump_deletion(
                    {entry.state_group}
                )
                if missing_state_group:
                    raise Exception(f"Missing state group: {entry.state_group}")
            elif entry.prev_group:
                # We only rely on the prev group when persisting the event if we
                # don't have an `entry.state_group`.
                missing_state_group = await self._state_deletion_store.check_state_groups_and_bump_deletion(
                    {entry.prev_group}
                )

                if missing_state_group:
                    # If we're missing the prev group then we can just clear the
                    # entries, and rely on `entry._state` (which must exist if
                    # `entry.state_group` is None)
                    entry.prev_group = None
                    entry.delta_ids = None

            state_group_before_event_prev_group = entry.prev_group
            deltas_to_state_group_before_event = entry.delta_ids
            state_ids_before_event = None

            # We make sure that we have a state group assigned to the state.
            if entry.state_group is None:
                # store_state_group requires us to have either a previous state group
                # (with deltas) or the complete state map. So, if we don't have a
                # previous state group, load the complete state map now.
                if state_group_before_event_prev_group is None:
                    state_ids_before_event = await entry.get_state(
                        self._state_storage_controller, StateFilter.all()
                    )

                state_group_before_event = (
                    await self._state_storage_controller.store_state_group(
                        event.event_id,
                        event.room_id,
                        prev_group=state_group_before_event_prev_group,
                        delta_ids=deltas_to_state_group_before_event,
                        current_state_ids=state_ids_before_event,
                    )
                )
                entry.set_state_group(state_group_before_event)
            else:
                state_group_before_event = entry.state_group

        #
        # now if it's not a state event, we're done
        #

        if not event.is_state():
            return UnpersistedEventContext(
                storage=self._storage_controllers,
                state_group_before_event=state_group_before_event,
                state_group_after_event=state_group_before_event,
                state_delta_due_to_event={},
                prev_group_for_state_group_before_event=state_group_before_event_prev_group,
                delta_ids_to_state_group_before_event=deltas_to_state_group_before_event,
                partial_state=partial_state,
                state_map_before_event=state_ids_before_event,
            )

        #
        # otherwise, we'll need to set up creating a new state group for after the event
        #

        key = (event.type, event.state_key)

        if state_ids_before_event is not None:
            replaces = state_ids_before_event.get(key)
        else:
            replaces_state_map = await entry.get_state(
                self._state_storage_controller, StateFilter.from_types([key])
            )
            replaces = replaces_state_map.get(key)

        if replaces and replaces != event.event_id:
            event.unsigned["replaces_state"] = replaces

        delta_ids = {key: event.event_id}

        return UnpersistedEventContext(
            storage=self._storage_controllers,
            state_group_before_event=state_group_before_event,
            state_group_after_event=None,
            state_delta_due_to_event=delta_ids,
            prev_group_for_state_group_before_event=state_group_before_event_prev_group,
            delta_ids_to_state_group_before_event=deltas_to_state_group_before_event,
            partial_state=partial_state,
            state_map_before_event=state_ids_before_event,
        )

    async def compute_event_context(
        self,
        event: EventBase,
        state_ids_before_event: StateMap[str] | None = None,
        partial_state: bool | None = None,
    ) -> EventContext:
        """Build an EventContext structure for a non-outlier event.

        (for an outlier, call EventContext.for_outlier directly)

        This works out what the current state should be for the event, and
        generates a new state group if necessary.

        Args:
            event:
            state_ids_before_event: The event ids of the state before the event if
                it can't be calculated from existing events. This is normally
                only specified when receiving an event from federation where we
                don't have the prev events, e.g. when backfilling.
            partial_state:
                `True` if `state_ids_before_event` is partial and omits non-critical
                membership events.
                `False` if `state_ids_before_event` is the full state.
                `None` when `state_ids_before_event` is not provided. In this case, the
                flag will be calculated based on `event`'s prev events.
            entry:
                A state cache entry for the resolved state across the prev events. We may
                have already calculated this, so if it's available pass it in
        Returns:
            The event context.

        Raises:
            RuntimeError if `state_ids_before_event` is not provided and one or more
                prev events are missing or outliers.
        """

        unpersisted_context = await self.calculate_context_info(
            event=event,
            state_ids_before_event=state_ids_before_event,
            partial_state=partial_state,
        )

        return await unpersisted_context.persist(event)

    @trace
    @measure_func()
    async def resolve_state_groups_for_events(
        self,
        room_id: str,
        event_ids: StrCollection,
        await_full_state: bool = True,
    ) -> _StateCacheEntry:
        """Given a list of event_ids this method fetches the state at each
        event, resolves conflicts between them and returns them.

        Args:
            room_id
            event_ids
            await_full_state: if true, will block if we do not yet have complete
               state at these events.

        Returns:
            The resolved state

        Raises:
            RuntimeError if we don't have a state group for one or more of the events
               (ie. they are outliers or unknown)
        """
        logger.debug("resolve_state_groups event_ids %s", event_ids)

        state_groups = await self._state_storage_controller.get_state_group_for_events(
            event_ids, await_full_state=await_full_state
        )

        state_group_ids = state_groups.values()

        # check if each event has same state group id, if so there's no state to resolve
        state_group_ids_set = set(state_group_ids)
        if len(state_group_ids_set) == 1:
            (state_group_id,) = state_group_ids_set
            (
                prev_group,
                delta_ids,
            ) = await self._state_storage_controller.get_state_group_delta(
                state_group_id
            )

            return _StateCacheEntry(
                state=None,
                state_group=state_group_id,
                prev_group=prev_group,
                delta_ids=delta_ids,
            )
        elif len(state_group_ids_set) == 0:
            return _StateCacheEntry(state={}, state_group=None)

        room_version = await self.store.get_room_version_id(room_id)

        state_to_resolve = await self._state_storage_controller.get_state_for_groups(
            state_group_ids_set
        )

        result = await self._state_resolution_handler.resolve_state_groups(
            room_id,
            room_version,
            state_to_resolve,
            None,
            state_res_store=StateResolutionStore(
                self.store, self._state_deletion_store
            ),
        )
        return result

    async def update_current_state(self, room_id: str) -> None:
        """Recalculates the current state for a room, and persists it.

        Raises:
            SynapseError(502): if all attempts to connect to the event persister worker
                fail
        """
        writer_instance = self._events_shard_config.get_instance(room_id)
        if writer_instance != self._instance_name:
            await self._update_current_state_client(
                instance_name=writer_instance,
                room_id=room_id,
            )
            return

        assert self._storage_controllers.persistence is not None
        await self._storage_controllers.persistence.update_current_state(room_id)


@attr.s(slots=True, auto_attribs=True)
class _StateResMetrics:
    """Keeps track of some usage metrics about state res."""

    # System and User CPU time, in seconds
    cpu_time: float = 0.0

    # time spent on database transactions (excluding scheduling time). This roughly
    # corresponds to the amount of work done on the db server, excluding event fetches.
    db_time: float = 0.0

    # number of events fetched from the db.
    db_events: int = 0


_biggest_room_by_cpu_counter = Counter(
    "synapse_state_res_cpu_for_biggest_room_seconds",
    "CPU time spent performing state resolution for the single most expensive "
    "room for state resolution",
    labelnames=[SERVER_NAME_LABEL],
)
_biggest_room_by_db_counter = Counter(
    "synapse_state_res_db_for_biggest_room_seconds",
    "Database time spent performing state resolution for the single most "
    "expensive room for state resolution",
    labelnames=[SERVER_NAME_LABEL],
)

_cpu_times = Histogram(
    "synapse_state_res_cpu_for_all_rooms_seconds",
    "CPU time (utime+stime) spent computing a single state resolution",
    labelnames=[SERVER_NAME_LABEL],
)
_db_times = Histogram(
    "synapse_state_res_db_for_all_rooms_seconds",
    "Database time spent computing a single state resolution",
    labelnames=[SERVER_NAME_LABEL],
)


class StateResolutionHandler:
    """Responsible for doing state conflict resolution.

    Note that the storage layer depends on this handler, so all functions must
    be storage-independent.
    """

    def __init__(self, hs: "HomeServer"):
        self.server_name = hs.hostname
        self.clock = hs.get_clock()

        self.resolve_linearizer = Linearizer(
            name="state_resolve_lock", clock=self.clock
        )

        # dict of set of event_ids -> _StateCacheEntry.
        self._state_cache: ExpiringCache[frozenset[int], _StateCacheEntry] = (
            ExpiringCache(
                cache_name="state_cache",
                server_name=self.server_name,
                hs=hs,
                clock=self.clock,
                max_len=100000,
                expiry_ms=EVICTION_TIMEOUT_SECONDS * 1000,
                iterable=True,
                reset_expiry_on_get=True,
            )
        )

        #
        # stuff for tracking time spent on state-res by room
        #

        # tracks the amount of work done on state res per room
        self._state_res_metrics: defaultdict[str, _StateResMetrics] = defaultdict(
            _StateResMetrics
        )

        self.clock.looping_call(self._report_metrics, Duration(minutes=2))

    async def resolve_state_groups(
        self,
        room_id: str,
        room_version: str,
        state_groups_ids: Mapping[int, StateMap[str]],
        event_map: dict[str, EventBase] | None,
        state_res_store: "StateResolutionStore",
    ) -> _StateCacheEntry:
        """Resolves conflicts between a set of state groups

        Always generates a new state group (unless we hit the cache), so should
        not be called for a single state group

        Args:
            room_id: room we are resolving for (used for logging and sanity checks)
            room_version: version of the room
            state_groups_ids:
                A map from state group id to the state in that state group
                (where 'state' is a map from state key to event id)

            event_map:
                a dict from event_id to event, for any events that we happen to
                have in flight (eg, those currently being persisted). This will be
                used as a starting point for finding the state we need; any missing
                events will be requested via state_res_store.

                If None, all events will be fetched via state_res_store.

            state_res_store

        Returns:
            The resolved state
        """
        group_names = frozenset(state_groups_ids.keys())

        async with self.resolve_linearizer.queue(group_names):
            cache = self._state_cache.get(group_names, None)
            if cache:
                # Check that the returned cache entry doesn't point to deleted
                # state groups.
                state_groups_to_check = set()
                if cache.state_group is not None:
                    state_groups_to_check.add(cache.state_group)

                if cache.prev_group is not None:
                    state_groups_to_check.add(cache.prev_group)

                missing_state_groups = await state_res_store.state_deletion_store.check_state_groups_and_bump_deletion(
                    state_groups_to_check
                )

                if not missing_state_groups:
                    return cache
                else:
                    # There are missing state groups, so let's remove the stale
                    # entry and continue as if it was a cache miss.
                    self._state_cache.pop(group_names, None)

            logger.info(
                "Resolving state for %s with groups %s",
                room_id,
                list(group_names),
            )

            # We double check that none of the state groups have been deleted.
            # They shouldn't be as all these state groups should be referenced.
            missing_state_groups = await state_res_store.state_deletion_store.check_state_groups_and_bump_deletion(
                group_names
            )
            if missing_state_groups:
                raise Exception(
                    f"State groups have been deleted: {shortstr(missing_state_groups)}"
                )

            state_groups_histogram.labels(
                **{SERVER_NAME_LABEL: self.server_name}
            ).observe(len(state_groups_ids))

            new_state = await self.resolve_events_with_store(
                room_id,
                room_version,
                list(state_groups_ids.values()),
                event_map=event_map,
                state_res_store=state_res_store,
            )

            # if the new state matches any of the input state groups, we can
            # use that state group again. Otherwise we will generate a state_id
            # which will be used as a cache key for future resolutions, but
            # not get persisted.

            with Measure(
                self.clock, name="state.create_group_ids", server_name=self.server_name
            ):
                cache = _make_state_cache_entry(new_state, state_groups_ids)

            self._state_cache[group_names] = cache

            return cache

    async def resolve_events_with_store(
        self,
        room_id: str,
        room_version: str,
        state_sets: Sequence[StateMap[str]],
        event_map: dict[str, EventBase] | None,
        state_res_store: "StateResolutionStore",
    ) -> StateMap[str]:
        """
        Args:
            room_id: the room we are working in

            room_version: Version of the room

            state_sets: List of dicts of (type, state_key) -> event_id,
                which are the different state groups to resolve.

            event_map:
                a dict from event_id to event, for any events that we happen to
                have in flight (eg, those currently being persisted). This will be
                used as a starting point for finding the state we need; any missing
                events will be requested via state_map_factory.

                If None, all events will be fetched via state_res_store.

            state_res_store: a place to fetch events from

        Returns:
            a map from (type, state_key) to event_id.
        """
        try:
            with Measure(
                self.clock, name="state._resolve_events", server_name=self.server_name
            ) as m:
                room_version_obj = KNOWN_ROOM_VERSIONS[room_version]
                if room_version_obj.state_res == StateResolutionVersions.V1:
                    return await v1.resolve_events_with_store(
                        room_id,
                        room_version_obj,
                        state_sets,
                        event_map,
                        state_res_store.get_events,
                    )
                else:
                    return await v2.resolve_events_with_store(
                        self.clock,
                        room_id,
                        room_version_obj,
                        state_sets,
                        event_map,
                        state_res_store,
                    )
        finally:
            self._record_state_res_metrics(room_id, m.get_resource_usage())

    def _record_state_res_metrics(
        self, room_id: str, rusage: ContextResourceUsage
    ) -> None:
        room_metrics = self._state_res_metrics[room_id]
        room_metrics.cpu_time += rusage.ru_utime + rusage.ru_stime
        room_metrics.db_time += rusage.db_txn_duration_sec
        room_metrics.db_events += rusage.evt_db_fetch_count

        _cpu_times.labels(**{SERVER_NAME_LABEL: self.server_name}).observe(
            rusage.ru_utime + rusage.ru_stime
        )
        _db_times.labels(**{SERVER_NAME_LABEL: self.server_name}).observe(
            rusage.db_txn_duration_sec
        )

    def _report_metrics(self) -> None:
        if not self._state_res_metrics:
            # no state res has happened since the last iteration: don't bother logging.
            return

        self._report_biggest(
            lambda i: i.cpu_time,
            "CPU time",
            _biggest_room_by_cpu_counter,
        )

        self._report_biggest(
            lambda i: i.db_time,
            "DB time",
            _biggest_room_by_db_counter,
        )

        self._state_res_metrics.clear()

    def _report_biggest(
        self,
        extract_key: Callable[[_StateResMetrics], Any],
        metric_name: str,
        prometheus_counter_metric: Counter,
    ) -> None:
        """Report metrics on the biggest rooms for state res

        Args:
            extract_key: a callable which, given a _StateResMetrics, extracts a single
                metric to sort by.
            metric_name: the name of the metric we have extracted, for the log line
            prometheus_counter_metric: a prometheus metric recording the sum of the
                the extracted metric
        """
        n_to_log = 10
        if not metrics_logger.isEnabledFor(logging.DEBUG):
            # only need the most expensive if we don't have debug logging, which
            # allows nlargest() to degrade to max()
            n_to_log = 1

        items = self._state_res_metrics.items()

        # log the N biggest rooms
        biggest: list[tuple[str, _StateResMetrics]] = heapq.nlargest(
            n_to_log, items, key=lambda i: extract_key(i[1])
        )
        metrics_logger.debug(
            "%i biggest rooms for state-res by %s: %s",
            len(biggest),
            metric_name,
            ["%s (%gs)" % (r, extract_key(m)) for (r, m) in biggest],
        )

        # report info on the single biggest to prometheus
        _, biggest_metrics = biggest[0]
        prometheus_counter_metric.labels(**{SERVER_NAME_LABEL: self.server_name}).inc(
            extract_key(biggest_metrics)
        )


def _make_state_cache_entry(
    new_state: StateMap[str], state_groups_ids: Mapping[int, StateMap[str]]
) -> _StateCacheEntry:
    """Given a resolved state, and a set of input state groups, pick one to base
    a new state group on (if any), and return an appropriately-constructed
    _StateCacheEntry.

    Args:
        new_state: resolved state map (mapping from (type, state_key) to event_id)

        state_groups_ids:
            map from state group id to the state in that state group (where
            'state' is a map from state key to event id)

    Returns:
        The cache entry.
    """
    # if the new state matches any of the input state groups, we can
    # use that state group again. Otherwise we will generate a state_id
    # which will be used as a cache key for future resolutions, but
    # not get persisted.

    # first look for exact matches
    new_state_event_ids = set(new_state.values())
    for sg, state in state_groups_ids.items():
        if len(new_state_event_ids) != len(state):
            continue

        old_state_event_ids = set(state.values())
        if new_state_event_ids == old_state_event_ids:
            # got an exact match.
            return _StateCacheEntry(state=None, state_group=sg)

    # TODO: We want to create a state group for this set of events, to
    # increase cache hits, but we need to make sure that it doesn't
    # end up as a prev_group without being added to the database

    # failing that, look for the closest match.
    prev_group = None
    delta_ids: StateMap[str] | None = None

    for old_group, old_state in state_groups_ids.items():
        if old_state.keys() - new_state.keys():
            # Currently we don't support deltas that remove keys from the state
            # map, so we have to ignore this group as a candidate to base the
            # new group on.
            continue

        n_delta_ids = {k: v for k, v in new_state.items() if old_state.get(k) != v}
        if not delta_ids or len(n_delta_ids) < len(delta_ids):
            prev_group = old_group
            delta_ids = n_delta_ids

    if prev_group is not None:
        # If we have a prev group and deltas then we can drop the new state from
        # the cache (to reduce memory usage).
        return _StateCacheEntry(
            state=None, state_group=None, prev_group=prev_group, delta_ids=delta_ids
        )
    else:
        return _StateCacheEntry(state=new_state, state_group=None)


@attr.s(slots=True, auto_attribs=True)
class StateResolutionStore:
    """Interface that allows state resolution algorithms to access the database
    in well defined way.
    """

    main_store: "DataStore"
    state_deletion_store: "StateDeletionDataStore"

    def get_events(
        self, event_ids: StrCollection, allow_rejected: bool = False
    ) -> Awaitable[dict[str, EventBase]]:
        """Get events from the database

        Args:
            event_ids: The event_ids of the events to fetch
            allow_rejected: If True return rejected events.

        Returns:
            An awaitable which resolves to a dict from event_id to event.
        """

        return self.main_store.get_events(
            event_ids,
            redact_behaviour=EventRedactBehaviour.as_is,
            get_prev_content=False,
            allow_rejected=allow_rejected,
        )

    def get_auth_chain_difference(
        self,
        room_id: str,
        state_sets: list[set[str]],
        conflicted_state: set[str] | None,
        additional_backwards_reachable_conflicted_events: set[str] | None,
    ) -> Awaitable[StateDifference]:
        """ "Given sets of state events figure out the auth chain difference (as
        per state res v2 algorithm).

        This is equivalent to fetching the full auth chain for each set of state
        and returning the events that don't appear in each and every auth
        chain.

        If conflicted_state is not None, calculate and return the conflicted sub-graph as per
        state res v2.1. The event IDs in the conflicted state MUST be a subset of the event IDs in
        state_sets.

        If additional_backwards_reachable_conflicted_events is set, the provided events are included
        when calculating the conflicted subgraph. This is primarily useful for calculating the
        subgraph across a combination of persisted and unpersisted events.

        Returns:
            information on the auth chain difference, and also the conflicted subgraph if
            conflicted_state is not None
        """

        return self.main_store.get_auth_chain_difference_extended(
            room_id,
            state_sets,
            conflicted_state,
            additional_backwards_reachable_conflicted_events,
        )



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

from os import path
from typing import Any

from synapse.config import ConfigError
from synapse.types import JsonDict

from ._base import Config


class ConsentConfig(Config):
    section = "consent"

    def __init__(self, *args: Any):
        super().__init__(*args)

        self.user_consent_version: str | None = None
        self.user_consent_template_dir: str | None = None
        self.user_consent_server_notice_content: JsonDict | None = None
        self.user_consent_server_notice_to_guests = False
        self.block_events_without_consent_error: str | None = None
        self.user_consent_at_registration = False
        self.user_consent_policy_name = "Privacy Policy"

    def read_config(self, config: JsonDict, **kwargs: Any) -> None:
        consent_config = config.get("user_consent")
        self.terms_template = self.read_template("terms.html")

        if consent_config is None:
            return
        self.user_consent_version = str(consent_config["version"])
        self.user_consent_template_dir = self.abspath(consent_config["template_dir"])
        if not isinstance(self.user_consent_template_dir, str) or not path.isdir(
            self.user_consent_template_dir
        ):
            raise ConfigError(
                "Could not find template directory '%s'"
                % (self.user_consent_template_dir,)
            )
        self.user_consent_server_notice_content = consent_config.get(
            "server_notice_content"
        )
        self.block_events_without_consent_error = consent_config.get(
            "block_events_error"
        )
        self.user_consent_server_notice_to_guests = bool(
            consent_config.get("send_server_notice_to_guests", False)
        )
        self.user_consent_at_registration = bool(
            consent_config.get("require_at_registration", False)
        )
        self.user_consent_policy_name = consent_config.get(
            "policy_name", "Privacy Policy"
        )



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2021 The Matrix.org Foundation C.I.C.
# Copyright 2014-2016 OpenMarket Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
import collections.abc
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Collection,
    Mapping,
    Match,
    MutableMapping,
)

import attr
from canonicaljson import encode_canonical_json

from synapse.api.constants import (
    CANONICALJSON_MAX_INT,
    CANONICALJSON_MIN_INT,
    MAX_PDU_SIZE,
    EventContentFields,
    EventTypes,
    RelationTypes,
)
from synapse.api.errors import Codes, SynapseError
from synapse.api.room_versions import RoomVersion
from synapse.logging.opentracing import SynapseTags, set_tag, trace
from synapse.types import JsonDict, Requester

from . import EventBase, FrozenEventV2, StrippedStateEvent, make_event_from_dict

if TYPE_CHECKING:
    from synapse.handlers.relations import BundledAggregations
    from synapse.server import HomeServer


# Split strings on "." but not "\." (or "\\\.").
SPLIT_FIELD_REGEX = re.compile(r"\\*\.")
# Find escaped characters, e.g. those with a \ in front of them.
ESCAPE_SEQUENCE_PATTERN = re.compile(r"\\(.)")


# Module API callback that allows adding fields to the unsigned section of
# events that are sent to clients.
ADD_EXTRA_FIELDS_TO_UNSIGNED_CLIENT_EVENT_CALLBACK = Callable[
    [EventBase], Awaitable[JsonDict]
]


def prune_event(event: EventBase) -> EventBase:
    """Returns a pruned version of the given event, which removes all keys we
    don't know about or think could potentially be dodgy.

    This is used when we "redact" an event. We want to remove all fields that
    the user has specified, but we do want to keep necessary information like
    type, state_key etc.
    """
    pruned_event_dict = prune_event_dict(event.room_version, event.get_dict())

    pruned_event = make_event_from_dict(
        pruned_event_dict, event.room_version, event.internal_metadata.get_dict()
    )

    # Copy the bits of `internal_metadata` that aren't returned by `get_dict`
    pruned_event.internal_metadata.stream_ordering = (
        event.internal_metadata.stream_ordering
    )
    pruned_event.internal_metadata.instance_name = event.internal_metadata.instance_name
    pruned_event.internal_metadata.outlier = event.internal_metadata.outlier

    # Mark the event as redacted
    pruned_event.internal_metadata.redacted = True

    return pruned_event


def clone_event(event: EventBase) -> EventBase:
    """Take a copy of the event.

    This is mostly useful because it does a *shallow* copy of the `unsigned` data,
    which means it can then be updated without corrupting the in-memory cache. Note that
    other properties of the event, such as `content`, are *not* (currently) copied here.
    """
    # XXX: We rely on at least one of `event.get_dict()` and `make_event_from_dict()`
    #   making a copy of `unsigned`. Currently, both do, though I don't really know why.
    #   Still, as long as they do, there's not much point doing yet another copy here.
    new_event = make_event_from_dict(
        event.get_dict(), event.room_version, event.internal_metadata.get_dict()
    )

    # Starting FrozenEventV2, the event ID is an (expensive) hash of the event. This is
    # lazily computed when we get the FrozenEventV2.event_id property, then cached in
    # _event_id field. Later FrozenEvent formats all inherit from FrozenEventV2, so we
    # can use the same logic here.
    if isinstance(event, FrozenEventV2) and isinstance(new_event, FrozenEventV2):
        # If we already pre-computed the event ID, use it.
        new_event._event_id = event._event_id

    # Copy the bits of `internal_metadata` that aren't returned by `get_dict`.
    new_event.internal_metadata.stream_ordering = (
        event.internal_metadata.stream_ordering
    )
    new_event.internal_metadata.instance_name = event.internal_metadata.instance_name
    new_event.internal_metadata.outlier = event.internal_metadata.outlier

    return new_event


def prune_event_dict(room_version: RoomVersion, event_dict: JsonDict) -> JsonDict:
    """Redacts the event_dict in the same way as `prune_event`, except it
    operates on dicts rather than event objects

    Returns:
        A copy of the pruned event dict
    """

    allowed_keys = [
        "event_id",
        "sender",
        "room_id",
        "hashes",
        "signatures",
        "content",
        "type",
        "state_key",
        "depth",
        "prev_events",
        "auth_events",
        "origin_server_ts",
    ]

    # Earlier room versions from had additional allowed keys.
    if not room_version.updated_redaction_rules:
        allowed_keys.extend(["prev_state", "membership", "origin"])

    event_type = event_dict["type"]

    new_content = {}

    def add_fields(*fields: str) -> None:
        for field in fields:
            if field in event_dict["content"]:
                new_content[field] = event_dict["content"][field]

    if event_type == EventTypes.Member:
        add_fields("membership")
        if room_version.restricted_join_rule_fix:
            add_fields(EventContentFields.AUTHORISING_USER)
        if room_version.updated_redaction_rules:
            # Preserve the signed field under third_party_invite.
            third_party_invite = event_dict["content"].get("third_party_invite")
            if isinstance(third_party_invite, collections.abc.Mapping):
                new_content["third_party_invite"] = {}
                if "signed" in third_party_invite:
                    new_content["third_party_invite"]["signed"] = third_party_invite[
                        "signed"
                    ]

    elif event_type == EventTypes.Create:
        if room_version.updated_redaction_rules:
            # MSC2176 rules state that create events cannot have their `content` redacted.
            new_content = event_dict["content"]
        if not room_version.implicit_room_creator:
            # Some room versions give meaning to `creator`
            add_fields("creator")
        if room_version.msc4291_room_ids_as_hashes:
            # room_id is not allowed on the create event as it's derived from the event ID
            allowed_keys.remove("room_id")

    elif event_type == EventTypes.JoinRules:
        add_fields("join_rule")
        if room_version.restricted_join_rule:
            add_fields("allow")
    elif event_type == EventTypes.PowerLevels:
        add_fields(
            "users",
            "users_default",
            "events",
            "events_default",
            "state_default",
            "ban",
            "kick",
            "redact",
        )

        if room_version.updated_redaction_rules:
            add_fields("invite")

    elif event_type == EventTypes.Aliases and room_version.special_case_aliases_auth:
        add_fields("aliases")
    elif event_type == EventTypes.RoomHistoryVisibility:
        add_fields("history_visibility")
    elif event_type == EventTypes.Redaction and room_version.updated_redaction_rules:
        add_fields("redacts")

    # Protect the rel_type and event_id fields under the m.relates_to field.
    if room_version.msc3389_relation_redactions:
        relates_to = event_dict["content"].get("m.relates_to")
        if isinstance(relates_to, collections.abc.Mapping):
            new_relates_to = {}
            for field in ("rel_type", "event_id"):
                if field in relates_to:
                    new_relates_to[field] = relates_to[field]
            # Only include a non-empty relates_to field.
            if new_relates_to:
                new_content["m.relates_to"] = new_relates_to

    allowed_fields = {k: v for k, v in event_dict.items() if k in allowed_keys}

    allowed_fields["content"] = new_content

    unsigned: JsonDict = {}
    allowed_fields["unsigned"] = unsigned

    event_unsigned = event_dict.get("unsigned", {})

    if "age_ts" in event_unsigned:
        unsigned["age_ts"] = event_unsigned["age_ts"]
    if "replaces_state" in event_unsigned:
        unsigned["replaces_state"] = event_unsigned["replaces_state"]

    return allowed_fields


def _copy_field(src: JsonDict, dst: JsonDict, field: list[str]) -> None:
    """Copy the field in 'src' to 'dst'.

    For example, if src={"foo":{"bar":5}} and dst={}, and field=["foo","bar"]
    then dst={"foo":{"bar":5}}.

    Args:
        src: The dict to read from.
        dst: The dict to modify.
        field: List of keys to drill down to in 'src'.
    """
    if len(field) == 0:  # this should be impossible
        return
    if len(field) == 1:  # common case e.g. 'origin_server_ts'
        if field[0] in src:
            dst[field[0]] = src[field[0]]
        return

    # Else is a nested field e.g. 'content.body'
    # Pop the last field as that's the key to move across and we need the
    # parent dict in order to access the data. Drill down to the right dict.
    key_to_move = field.pop(-1)
    sub_dict = src
    for sub_field in field:  # e.g. sub_field => "content"
        if sub_field in sub_dict and isinstance(
            sub_dict[sub_field], collections.abc.Mapping
        ):
            sub_dict = sub_dict[sub_field]
        else:
            return

    if key_to_move not in sub_dict:
        return

    # Insert the key into the output dictionary, creating nested objects
    # as required. We couldn't do this any earlier or else we'd need to delete
    # the empty objects if the key didn't exist.
    sub_out_dict = dst
    for sub_field in field:
        sub_out_dict = sub_out_dict.setdefault(sub_field, {})
    sub_out_dict[key_to_move] = sub_dict[key_to_move]


def _escape_slash(m: Match[str]) -> str:
    """
    Replacement function; replace a backslash-backslash or backslash-dot with the
    second character. Leaves any other string alone.
    """
    if m.group(1) in ("\\", "."):
        return m.group(1)
    return m.group(0)


def _split_field(field: str) -> list[str]:
    """
    Splits strings on unescaped dots and removes escaping.

    Args:
        field: A string representing a path to a field.

    Returns:
        A list of nested fields to traverse.
    """

    # Convert the field and remove escaping:
    #
    # 1. "content.body.thing\.with\.dots"
    # 2. ["content", "body", "thing\.with\.dots"]
    # 3. ["content", "body", "thing.with.dots"]

    # Find all dots (and their preceding backslashes). If the dot is unescaped
    # then emit a new field part.
    result = []
    prev_start = 0
    for match in SPLIT_FIELD_REGEX.finditer(field):
        # If the match is an *even* number of characters than the dot was escaped.
        if len(match.group()) % 2 == 0:
            continue

        # Add a new part (up to the dot, exclusive) after escaping.
        result.append(
            ESCAPE_SEQUENCE_PATTERN.sub(
                _escape_slash, field[prev_start : match.end() - 1]
            )
        )
        prev_start = match.end()

    # Add any part of the field after the last unescaped dot. (Note that if the
    # character is a dot this correctly adds a blank string.)
    result.append(re.sub(r"\\(.)", _escape_slash, field[prev_start:]))

    return result


def only_fields(dictionary: JsonDict, fields: list[str]) -> JsonDict:
    """Return a new dict with only the fields in 'dictionary' which are present
    in 'fields'.

    If there are no event fields specified then all fields are included.
    The entries may include '.' characters to indicate sub-fields.
    So ['content.body'] will include the 'body' field of the 'content' object.
    A literal '.' or '\' character in a field name may be escaped using a '\'.

    Args:
        dictionary: The dictionary to read from.
        fields: A list of fields to copy over. Only shallow refs are
        taken.
    Returns:
        A new dictionary with only the given fields. If fields was empty,
        the same dictionary is returned.
    """
    if len(fields) == 0:
        return dictionary

    # for each field, convert it:
    # ["content.body.thing\.with\.dots"] => [["content", "body", "thing\.with\.dots"]]
    split_fields = [_split_field(f) for f in fields]

    output: JsonDict = {}
    for field_array in split_fields:
        _copy_field(dictionary, output, field_array)
    return output


def format_event_raw(d: JsonDict) -> JsonDict:
    return d


def format_event_for_client_v1(d: JsonDict) -> JsonDict:
    d = format_event_for_client_v2(d)

    sender = d.get("sender")
    if sender is not None:
        d["user_id"] = sender

    copy_keys = (
        "age",
        "redacted_because",
        "replaces_state",
        "prev_content",
        "invite_room_state",
        "knock_room_state",
    )
    for key in copy_keys:
        if key in d["unsigned"]:
            d[key] = d["unsigned"][key]

    return d


def format_event_for_client_v2(d: JsonDict) -> JsonDict:
    drop_keys = (
        "auth_events",
        "prev_events",
        "hashes",
        "signatures",
        "depth",
        "origin",
        "prev_state",
    )
    for key in drop_keys:
        d.pop(key, None)
    return d


def format_event_for_client_v2_without_room_id(d: JsonDict) -> JsonDict:
    d = format_event_for_client_v2(d)
    d.pop("room_id", None)
    return d


@attr.s(slots=True, frozen=True, auto_attribs=True)
class SerializeEventConfig:
    as_client_event: bool = True
    # Function to convert from federation format to client format
    event_format: Callable[[JsonDict], JsonDict] = format_event_for_client_v1
    # The entity that requested the event. This is used to determine whether to include
    # the transaction_id and delay_id in the unsigned section of the event.
    requester: Requester | None = None
    # List of event fields to include. If empty, all fields will be returned.
    only_event_fields: list[str] | None = None
    # Some events can have stripped room state stored in the `unsigned` field.
    # This is required for invite and knock functionality. If this option is
    # False, that state will be removed from the event before it is returned.
    # Otherwise, it will be kept.
    include_stripped_room_state: bool = False
    # When True, sets unsigned fields to help clients identify events which
    # only server admins can see through other configuration. For example,
    # whether an event was soft failed by the server.
    include_admin_metadata: bool = False


_DEFAULT_SERIALIZE_EVENT_CONFIG = SerializeEventConfig()


def make_config_for_admin(existing: SerializeEventConfig) -> SerializeEventConfig:
    # Set the options which are only available to server admins,
    # and copy the rest.
    return attr.evolve(existing, include_admin_metadata=True)


def serialize_event(
    e: JsonDict | EventBase,
    time_now_ms: int,
    *,
    config: SerializeEventConfig = _DEFAULT_SERIALIZE_EVENT_CONFIG,
) -> JsonDict:
    """Serialize event for clients

    Args:
        e
        time_now_ms
        config: Event serialization config

    Returns:
        The serialized event dictionary.
    """

    # FIXME(erikj): To handle the case of presence events and the like
    if not isinstance(e, EventBase):
        return e

    time_now_ms = int(time_now_ms)

    # Should this strip out None's?
    d = dict(e.get_dict().items())

    d["event_id"] = e.event_id

    if "age_ts" in d["unsigned"]:
        d["unsigned"]["age"] = time_now_ms - d["unsigned"]["age_ts"]
        del d["unsigned"]["age_ts"]

    if "redacted_because" in e.unsigned:
        d["unsigned"]["redacted_because"] = serialize_event(
            e.unsigned["redacted_because"],
            time_now_ms,
            config=config,
        )

    # If we have applicable fields saved in the internal_metadata, include them in the
    # unsigned section of the event if the event was sent by the same session (or when
    # appropriate, just the same sender) as the one requesting the event.
    if config.requester is not None and config.requester.user.to_string() == e.sender:
        txn_id: str | None = getattr(e.internal_metadata, "txn_id", None)
        if txn_id is not None:
            # Some events do not have the device ID stored in the internal metadata,
            # this includes old events as well as those created by appservice, guests,
            # or with tokens minted with the admin API. For those events, fallback
            # to using the access token instead.
            event_device_id: str | None = getattr(
                e.internal_metadata, "device_id", None
            )
            if event_device_id is not None:
                if event_device_id == config.requester.device_id:
                    d["unsigned"]["transaction_id"] = txn_id

            else:
                # Fallback behaviour: only include the transaction ID if the event
                # was sent from the same access token.
                #
                # For regular users, the access token ID can be used to determine this.
                # This includes access tokens minted with the admin API.
                #
                # For guests and appservice users, we can't check the access token ID
                # so assume it is the same session.
                event_token_id: int | None = getattr(
                    e.internal_metadata, "token_id", None
                )
                if (
                    (
                        event_token_id is not None
                        and config.requester.access_token_id is not None
                        and event_token_id == config.requester.access_token_id
                    )
                    or config.requester.is_guest
                    or config.requester.app_service
                ):
                    d["unsigned"]["transaction_id"] = txn_id

        delay_id: str | None = getattr(e.internal_metadata, "delay_id", None)
        if delay_id is not None:
            d["unsigned"]["org.matrix.msc4140.delay_id"] = delay_id

    # invite_room_state and knock_room_state are a list of stripped room state events
    # that are meant to provide metadata about a room to an invitee/knocker. They are
    # intended to only be included in specific circumstances, such as down sync, and
    # should not be included in any other case.
    if not config.include_stripped_room_state:
        d["unsigned"].pop("invite_room_state", None)
        d["unsigned"].pop("knock_room_state", None)

    if config.as_client_event:
        d = config.event_format(d)

    # Ensure the room_id field is set for create events in MSC4291 rooms
    if e.type == EventTypes.Create and e.room_version.msc4291_room_ids_as_hashes:
        d["room_id"] = e.room_id

    # If the event is a redaction, the field with the redacted event ID appears
    # in a different location depending on the room version. e.redacts handles
    # fetching from the proper location; copy it to the other location for forwards-
    # and backwards-compatibility with clients.
    if e.type == EventTypes.Redaction and e.redacts is not None:
        if e.room_version.updated_redaction_rules:
            d["redacts"] = e.redacts
        else:
            d["content"] = dict(d["content"])
            d["content"]["redacts"] = e.redacts

    if config.include_admin_metadata:
        if e.internal_metadata.is_soft_failed():
            d["unsigned"]["io.element.synapse.soft_failed"] = True
        if e.internal_metadata.policy_server_spammy:
            d["unsigned"]["io.element.synapse.policy_server_spammy"] = True

    only_event_fields = config.only_event_fields
    if only_event_fields:
        if not isinstance(only_event_fields, list) or not all(
            isinstance(f, str) for f in only_event_fields
        ):
            raise TypeError("only_event_fields must be a list of strings")
        d = only_fields(d, only_event_fields)

    return d


class EventClientSerializer:
    """Serializes events that are to be sent to clients.

    This is used for bundling extra information with any events to be sent to
    clients.
    """

    def __init__(self, hs: "HomeServer") -> None:
        self._store = hs.get_datastores().main
        self._auth = hs.get_auth()
        self._add_extra_fields_to_unsigned_client_event_callbacks: list[
            ADD_EXTRA_FIELDS_TO_UNSIGNED_CLIENT_EVENT_CALLBACK
        ] = []

    async def serialize_event(
        self,
        event: JsonDict | EventBase,
        time_now: int,
        *,
        config: SerializeEventConfig = _DEFAULT_SERIALIZE_EVENT_CONFIG,
        bundle_aggregations: dict[str, "BundledAggregations"] | None = None,
    ) -> JsonDict:
        """Serializes a single event.

        Args:
            event: The event being serialized.
            time_now: The current time in milliseconds
            config: Event serialization config
            bundle_aggregations: A map from event_id to the aggregations to be bundled
               into the event.

        Returns:
            The serialized event
        """
        # To handle the case of presence events and the like
        if not isinstance(event, EventBase):
            return event

        # Force-enable server admin metadata because the only time an event with
        # relevant metadata will be when the admin requested it via their admin
        # client config account data. Also, it's "just" some `unsigned` fields, so
        # shouldn't cause much in terms of problems to downstream consumers.
        if config.requester is not None and await self._auth.is_server_admin(
            config.requester
        ):
            config = make_config_for_admin(config)

        serialized_event = serialize_event(event, time_now, config=config)

        new_unsigned = {}
        for callback in self._add_extra_fields_to_unsigned_client_event_callbacks:
            u = await callback(event)
            new_unsigned.update(u)

        if new_unsigned:
            # We do the `update` this way round so that modules can't clobber
            # existing fields.
            new_unsigned.update(serialized_event["unsigned"])
            serialized_event["unsigned"] = new_unsigned

        # Check if there are any bundled aggregations to include with the event.
        if bundle_aggregations:
            if event.event_id in bundle_aggregations:
                await self._inject_bundled_aggregations(
                    event,
                    time_now,
                    config,
                    bundle_aggregations,
                    serialized_event,
                )

        return serialized_event

    async def _inject_bundled_aggregations(
        self,
        event: EventBase,
        time_now: int,
        config: SerializeEventConfig,
        bundled_aggregations: dict[str, "BundledAggregations"],
        serialized_event: JsonDict,
    ) -> None:
        """Potentially injects bundled aggregations into the unsigned portion of the serialized event.

        Args:
            event: The event being serialized.
            time_now: The current time in milliseconds
            config: Event serialization config
            bundled_aggregations: Bundled aggregations to be injected.
                A map from event_id to aggregation data. Must contain at least an
                entry for `event`.

                While serializing the bundled aggregations this map may be searched
                again for additional events in a recursive manner.
            serialized_event: The serialized event which may be modified.
        """

        # We have already checked that aggregations exist for this event.
        event_aggregations = bundled_aggregations[event.event_id]

        # The JSON dictionary to be added under the unsigned property of the event
        # being serialized.
        serialized_aggregations = {}

        if event_aggregations.references:
            serialized_aggregations[RelationTypes.REFERENCE] = (
                event_aggregations.references
            )

        if event_aggregations.replace:
            # Include information about it in the relations dict.
            #
            # Matrix spec v1.5 (https://spec.matrix.org/v1.5/client-server-api/#server-side-aggregation-of-mreplace-relationships)
            # said that we should only include the `event_id`, `origin_server_ts` and
            # `sender` of the edit; however MSC3925 proposes extending it to the whole
            # of the edit, which is what we do here.
            serialized_aggregations[RelationTypes.REPLACE] = await self.serialize_event(
                event_aggregations.replace,
                time_now,
                config=config,
            )

        # Include any threaded replies to this event.
        if event_aggregations.thread:
            thread = event_aggregations.thread

            serialized_latest_event = await self.serialize_event(
                thread.latest_event,
                time_now,
                config=config,
                bundle_aggregations=bundled_aggregations,
            )

            thread_summary = {
                "latest_event": serialized_latest_event,
                "count": thread.count,
                "current_user_participated": thread.current_user_participated,
            }
            serialized_aggregations[RelationTypes.THREAD] = thread_summary

        # Include the bundled aggregations in the event.
        if serialized_aggregations:
            # There is likely already an "unsigned" field, but a filter might
            # have stripped it off (via the event_fields option). The server is
            # allowed to return additional fields, so add it back.
            serialized_event.setdefault("unsigned", {}).setdefault(
                "m.relations", {}
            ).update(serialized_aggregations)

    @trace
    async def serialize_events(
        self,
        events: Collection[JsonDict | EventBase],
        time_now: int,
        *,
        config: SerializeEventConfig = _DEFAULT_SERIALIZE_EVENT_CONFIG,
        bundle_aggregations: dict[str, "BundledAggregations"] | None = None,
    ) -> list[JsonDict]:
        """Serializes multiple events.

        Args:
            event
            time_now: The current time in milliseconds
            config: Event serialization config
            bundle_aggregations: Whether to include the bundled aggregations for this
                event. Only applies to non-state events. (State events never include
                bundled aggregations.)

        Returns:
            The list of serialized events
        """
        set_tag(
            SynapseTags.FUNC_ARG_PREFIX + "events.length",
            str(len(events)),
        )

        return [
            await self.serialize_event(
                event,
                time_now,
                config=config,
                bundle_aggregations=bundle_aggregations,
            )
            for event in events
        ]

    def register_add_extra_fields_to_unsigned_client_event_callback(
        self, callback: ADD_EXTRA_FIELDS_TO_UNSIGNED_CLIENT_EVENT_CALLBACK
    ) -> None:
        """Register a callback that returns additions to the unsigned section of
        serialized events.
        """
        self._add_extra_fields_to_unsigned_client_event_callbacks.append(callback)


_PowerLevel = str | int
PowerLevelsContent = Mapping[str, _PowerLevel | Mapping[str, _PowerLevel]]


def copy_and_fixup_power_levels_contents(
    old_power_levels: PowerLevelsContent,
) -> dict[str, int | dict[str, int]]:
    """Copy the content of a power_levels event, unfreezing immutabledicts along the way.

    We accept as input power level values which are strings, provided they represent an
    integer, e.g. `"`100"` instead of 100. Such strings are converted to integers
    in the returned dictionary (hence "fixup" in the function name).

    Note that future room versions will outlaw such stringy power levels (see
    https://github.com/matrix-org/matrix-spec/issues/853).

    Raises:
        TypeError if the input does not look like a valid power levels event content
    """
    if not isinstance(old_power_levels, collections.abc.Mapping):
        raise TypeError("Not a valid power-levels content: %r" % (old_power_levels,))

    power_levels: dict[str, int | dict[str, int]] = {}

    for k, v in old_power_levels.items():
        if isinstance(v, collections.abc.Mapping):
            h: dict[str, int] = {}
            power_levels[k] = h
            for k1, v1 in v.items():
                _copy_power_level_value_as_integer(v1, h, k1)

        else:
            _copy_power_level_value_as_integer(v, power_levels, k)

    return power_levels


def _copy_power_level_value_as_integer(
    old_value: object,
    power_levels: MutableMapping[str, Any],
    key: str,
) -> None:
    """Set `power_levels[key]` to the integer represented by `old_value`.

    :raises TypeError: if `old_value` is neither an integer nor a base-10 string
        representation of an integer.
    """
    if type(old_value) is int:  # noqa: E721
        power_levels[key] = old_value
        return

    if isinstance(old_value, str):
        try:
            parsed_value = int(old_value, base=10)
        except ValueError:
            # Fall through to the final TypeError.
            pass
        else:
            power_levels[key] = parsed_value
            return

    raise TypeError(f"Invalid power_levels value for {key}: {old_value}")


def validate_canonicaljson(value: Any) -> None:
    """
    Ensure that the JSON object is valid according to the rules of canonical JSON.

    See the appendix section 3.1: Canonical JSON.

    This rejects JSON that has:
    * An integer outside the range of [-2 ^ 53 + 1, 2 ^ 53 - 1]
    * Floats
    * NaN, Infinity, -Infinity
    """
    if type(value) is int:  # noqa: E721
        if value < CANONICALJSON_MIN_INT or CANONICALJSON_MAX_INT < value:
            raise SynapseError(400, "JSON integer out of range", Codes.BAD_JSON)

    elif isinstance(value, float):
        # Note that Infinity, -Infinity, and NaN are also considered floats.
        raise SynapseError(400, "Bad JSON value: float", Codes.BAD_JSON)

    elif isinstance(value, collections.abc.Mapping):
        for v in value.values():
            validate_canonicaljson(v)

    elif isinstance(value, (list, tuple)):
        for i in value:
            validate_canonicaljson(i)

    elif not isinstance(value, (bool, str)) and value is not None:
        # Other potential JSON values (bool, None, str) are safe.
        raise SynapseError(400, "Unknown JSON value", Codes.BAD_JSON)


def maybe_upsert_event_field(
    event: EventBase, container: JsonDict, key: str, value: object
) -> bool:
    """Upsert an event field, but only if this doesn't make the event too large.

    Returns true iff the upsert took place.
    """
    if key in container:
        old_value: object = container[key]
        container[key] = value
        # NB: here and below, we assume that passing a non-None `time_now` argument to
        # get_pdu_json doesn't increase the size of the encoded result.
        upsert_okay = len(encode_canonical_json(event.get_pdu_json())) <= MAX_PDU_SIZE
        if not upsert_okay:
            container[key] = old_value
    else:
        container[key] = value
        upsert_okay = len(encode_canonical_json(event.get_pdu_json())) <= MAX_PDU_SIZE
        if not upsert_okay:
            del container[key]

    return upsert_okay


def strip_event(event: EventBase) -> JsonDict:
    """
    Used for "stripped state" events which provide a simplified view of the state of a
    room intended to help a potential joiner identify the room (relevant when the user
    is invited or knocked).

    Stripped state events can only have the `sender`, `type`, `state_key` and `content`
    properties present.
    """
    # MSC4311: Ensure the create event is available on invites and knocks.
    # TODO: Implement the rest of MSC4311
    if (
        event.room_version.msc4291_room_ids_as_hashes
        and event.type == EventTypes.Create
        and event.get_state_key() == ""
    ):
        return event.get_pdu_json()

    return {
        "type": event.type,
        "state_key": event.state_key,
        "content": event.content,
        "sender": event.sender,
    }


def parse_stripped_state_event(raw_stripped_event: Any) -> StrippedStateEvent | None:
    """
    Given a raw value from an event's `unsigned` field, attempt to parse it into a
    `StrippedStateEvent`.
    """
    if isinstance(raw_stripped_event, dict):
        # All of these fields are required
        type = raw_stripped_event.get("type")
        state_key = raw_stripped_event.get("state_key")
        sender = raw_stripped_event.get("sender")
        content = raw_stripped_event.get("content")
        if (
            isinstance(type, str)
            and isinstance(state_key, str)
            and isinstance(sender, str)
            and isinstance(content, dict)
        ):
            return StrippedStateEvent(
                type=type,
                state_key=state_key,
                sender=sender,
                content=content,
            )

    return None



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2014-2016 OpenMarket Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from synapse.types import StrCollection, UserID

# The key, this is either a stream token or int.
K = TypeVar("K")
# The return type.
R = TypeVar("R")


class EventSource(ABC, Generic[K, R]):
    @abstractmethod
    async def get_new_events(
        self,
        user: UserID,
        from_key: K,
        limit: int,
        room_ids: StrCollection,
        is_guest: bool,
        explicit_room_id: str | None = None,
    ) -> tuple[list[R], K]:
        raise NotImplementedError()



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2017 Vector Creations Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

"""Defines all the valid streams that clients can subscribe to, and the format
of the rows returned by each stream.

Each stream is defined by the following information:

    stream name:        The name of the stream
    row type:           The type that is used to serialise/deserialse the row
    current_token:      The function that returns the current token for the stream
    update_function:    The function that returns a list of updates between two tokens
"""

from synapse.replication.tcp.streams._base import (
    AccountDataStream,
    BackfillStream,
    CachesStream,
    DeviceListsStream,
    PresenceFederationStream,
    PresenceStream,
    PushersStream,
    PushRulesStream,
    ReceiptsStream,
    StickyEventsStream,
    Stream,
    ThreadSubscriptionsStream,
    ToDeviceStream,
    TypingStream,
)
from synapse.replication.tcp.streams.events import EventsStream
from synapse.replication.tcp.streams.federation import FederationStream
from synapse.replication.tcp.streams.partial_state import (
    UnPartialStatedEventStream,
    UnPartialStatedRoomStream,
)

STREAMS_MAP = {
    stream.NAME: stream
    for stream in (
        EventsStream,
        BackfillStream,
        PresenceStream,
        PresenceFederationStream,
        TypingStream,
        ReceiptsStream,
        PushRulesStream,
        PushersStream,
        CachesStream,
        DeviceListsStream,
        ToDeviceStream,
        FederationStream,
        AccountDataStream,
        StickyEventsStream,
        ThreadSubscriptionsStream,
        UnPartialStatedRoomStream,
        UnPartialStatedEventStream,
    )
}

__all__ = [
    "STREAMS_MAP",
    "Stream",
    "EventsStream",
    "BackfillStream",
    "PresenceStream",
    "PresenceFederationStream",
    "TypingStream",
    "ReceiptsStream",
    "PushRulesStream",
    "PushersStream",
    "CachesStream",
    "DeviceListsStream",
    "ToDeviceStream",
    "FederationStream",
    "AccountDataStream",
    "StickyEventsStream",
    "ThreadSubscriptionsStream",
    "UnPartialStatedRoomStream",
    "UnPartialStatedEventStream",
]



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright (C) 2025 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#

import logging
from typing import TYPE_CHECKING, Awaitable, Callable

import attr

from synapse.util.async_helpers import delay_cancellation
from synapse.util.metrics import Measure

if TYPE_CHECKING:
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True)
class RatelimitOverride:
    """Represents a ratelimit being overridden."""

    per_second: float
    """The number of actions that can be performed in a second. `0.0` means that ratelimiting is disabled."""
    burst_count: int
    """How many actions that can be performed before being limited."""


GET_RATELIMIT_OVERRIDE_FOR_USER_CALLBACK = Callable[
    [str, str], Awaitable[RatelimitOverride | None]
]


class RatelimitModuleApiCallbacks:
    def __init__(self, hs: "HomeServer") -> None:
        self.server_name = hs.hostname
        self.clock = hs.get_clock()
        self._get_ratelimit_override_for_user_callbacks: list[
            GET_RATELIMIT_OVERRIDE_FOR_USER_CALLBACK
        ] = []

    def register_callbacks(
        self,
        get_ratelimit_override_for_user: GET_RATELIMIT_OVERRIDE_FOR_USER_CALLBACK
        | None = None,
    ) -> None:
        """Register callbacks from module for each hook."""
        if get_ratelimit_override_for_user is not None:
            self._get_ratelimit_override_for_user_callbacks.append(
                get_ratelimit_override_for_user
            )

    async def get_ratelimit_override_for_user(
        self, user_id: str, limiter_name: str
    ) -> RatelimitOverride | None:
        for callback in self._get_ratelimit_override_for_user_callbacks:
            with Measure(
                self.clock,
                name=f"{callback.__module__}.{callback.__qualname__}",
                server_name=self.server_name,
            ):
                res: RatelimitOverride | None = await delay_cancellation(
                    callback(user_id, limiter_name)
                )
            if res:
                return res

        return None
